#!/usr/bin/env node
import { command, cli } from 'cleye';
import _ from 'lodash';
import { select, text, confirm, outro, intro, group, cancel } from '@clack/prompts';
import { dim, red, green, blue, lightGreen, bgYellow, black } from 'kolorist';
import dedent from 'dedent';
import { promisify } from 'node:util';
import { exec, spawn } from 'node:child_process';

var name = "commitizen-cli";
var version = "1.3.1";
var author = {
	name: "iamando",
	email: "iamando@tuta.io"
};

const PACKAGE_NAME = name;
const VERSION = version;
const AUTHOR = author;
const CANCELED_OP_MSG = "Operation Canceled";

const type = () => select({
  message: "Choose commit type",
  options: [
    { label: "Feature", value: "feature" },
    { label: "Bugfix", value: "bugfix" },
    { label: "Hotfix", value: "hotfix" },
    { label: "Chore", value: "chore" },
    { label: "Epic", value: "epic" },
    { label: "Design", value: "design" },
    { label: "Experiment", value: "experiment" },
    { label: "Documentation", value: "documentation" }
  ]
});
const message = () => text({
  message: "Insert commit message",
  validate: (value) => {
    if (value.length === 0)
      return "Commit message required";
  }
});
const ticket = async () => {
  const hasTicket = await confirm({
    message: "Has Ticket ?",
    initialValue: false
  });
  if (!hasTicket)
    return;
  text({
    message: "Insert your Ticket name",
    validate: (value) => {
      if (value.length === 0)
        return "Ticket name required";
    }
  });
};

const log = ({
  type,
  msg,
  isConsole = true,
  newLine = true
}) => {
  switch (type) {
    case "info":
      return isConsole ? console.info(`${newLine ? "\n" : ""} ${blue("\u2754")} ${msg}`) : `${newLine ? "\n" : ""} ${blue("\u2754")} ${msg}`;
    case "success":
      return isConsole ? console.log(`${newLine ? "\n" : ""} ${green("\u2714")} ${msg}`) : `${newLine ? "\n" : ""} ${green("\u2714")} ${msg}`;
    case "error":
      return isConsole ? console.error(` ${newLine ? "\n" : ""} ${red("\u274C")} ${msg}`) : `${newLine ? "\n" : ""} ${red("\u274C")} ${msg}`;
    default:
      return isConsole ? console.log(`${newLine ? "\n" : ""} ${dim("\u2755")} ${msg}`) : `${newLine ? "\n" : ""} ${dim("\u2755")} ${msg}`;
  }
};

class CliError extends Error {
}
const indent = " ".repeat(4);
const handleCliError = (error) => {
  if (error instanceof Error && !(error instanceof CliError)) {
    if (error.stack) {
      log({
        type: "error",
        msg: error.stack.split("\n").slice(1).join("\n")
      });
    }
    log({
      type: "error",
      msg: `
${indent}${dim(`${PACKAGE_NAME} v${VERSION}`)}`
    });
    log({
      type: "error",
      msg: `
${indent}Please open a Bug report with the information above:`
    });
    log({
      type: "error",
      msg: `${indent}https://github.com/${AUTHOR.name}/${PACKAGE_NAME}/issues/new`
    });
  }
};

const formatCommitWithEmojiByType = ({
  type,
  message,
  ticket
}) => {
  const emojiByType = {
    feature: "\u2728",
    bug: "\u{1F41B}",
    hotfix: "\u{1F691}",
    chore: "\u{1F6E0}\uFE0F",
    epic: "\u{1F4CC}",
    design: "\u{1F3A8}",
    experiment: "\u{1F9EA}",
    documentation: "\u{1F4DD}"
  };
  let commit;
  if (!_.isEmpty(ticket)) {
    commit = `${emojiByType[type]} ${type}(${ticket}): ${message}`;
  } else {
    commit = `${emojiByType[type]} ${type}: ${message}`;
  }
  return commit;
};

const execa = promisify(exec);
const gitStatus = async () => {
  const { stdout: stdoutStatus, stderr: stderrStatus } = await execa(
    "git status"
  );
  return { stdoutStatus, stderrStatus };
};
const gitAdd = async () => {
  const { stdout: stdoutAdd, stderr: stderrAdd } = await execa("git add .");
  return { stdoutAdd, stderrAdd };
};
const gitCommit = async ({ commit }) => {
  const { stdout: stdoutCommit, stderr: stderrCommit } = await execa(
    `git commit -m "${commit}"`
  );
  return { stdoutCommit, stderrCommit };
};
const isTreeClean = async () => {
  const { stdoutStatus, stderrStatus } = await gitStatus();
  if (stderrStatus)
    throw new CliError(`An error occured: ${stderrStatus}`);
  if (stdoutStatus.includes("nothing to commit, working tree clean")) {
    outro(lightGreen("Nothing to commit, working tree clean \u{1F9F9}"));
    process.exit(1);
  }
};

const commiter = async () => {
  intro(bgYellow(black("Commitizen CLI")));
  await isTreeClean();
  const values = await group(
    {
      type: () => type(),
      message: () => message(),
      ticket: () => ticket()
    },
    {
      onCancel: () => {
        cancel(CANCELED_OP_MSG);
        process.exit(0);
      }
    }
  );
  try {
    let commit;
    const useEmoji = await confirm({
      message: "Use emoji on commit type ?",
      initialValue: false
    });
    if (useEmoji) {
      commit = formatCommitWithEmojiByType({
        type: values.type,
        message: values.message,
        ticket: values.ticket
      });
    } else {
      if (!_.isEmpty(values.ticket)) {
        commit = `${values.type}(${values.ticket}): ${values.message}`;
      } else {
        commit = `${values.type}: ${values.message}`;
      }
    }
    const { stdoutStatus, stderrStatus } = await gitStatus();
    if (stderrStatus)
      throw new CliError(`An error occured: ${stderrStatus}`);
    if (stdoutStatus.includes("no changes added to commit") || stdoutStatus.includes(
      "nothing added to commit but untracked files present"
    )) {
      const type2 = stdoutStatus.includes("no changes added to commit") ? "modified" : "untracked";
      const addStagedFiles = await confirm({
        message: `No changes added to commit, would you like to add ${type2} files ?`,
        initialValue: true
      });
      if (addStagedFiles) {
        const { stderrAdd } = await gitAdd();
        if (stderrAdd)
          throw new CliError(`An error occured: ${stderrAdd}`);
        const { stderrCommit: stderrCommit2 } = await gitCommit({ commit: commit.trim() });
        if (stderrCommit2)
          throw new CliError(`An error occured: ${stderrCommit2}`);
        outro(dedent`
        You're all set 🎉

        use "git push" to publish your local commits
      `);
      }
      return;
    }
    const { stderrCommit } = await gitCommit({ commit: commit.trim() });
    if (stderrCommit)
      throw new CliError(`An error occured: ${stderrCommit}`);
    outro(dedent`
    You're all set 🎉

    use "git push" to publish your local commits
  `);
  } catch (err) {
    log({ type: "error", msg: err.message });
    handleCliError(err);
    process.exit(1);
  }
};

var update = command(
  {
    name: "update",
    description: "Update Commitizen CLI to the latest version"
  },
  async () => {
    const cmd = "npm update -g commitizen-cli";
    console.log(dim(`Running: ${cmd}`));
    const update = spawn(cmd, {
      stdio: "inherit",
      shell: process.env.SHELL || true
    });
    update.stderr?.on("error", () => {
      throw new CliError("An error occured on updating package");
    });
    update.on("close", (code) => {
      if (code !== 0) {
        console.log(`update process exited with code ${code}`);
      }
    });
  }
);

cli(
  {
    name: PACKAGE_NAME,
    version: VERSION,
    commands: [update]
  },
  () => {
    commiter().catch((err) => {
      log({ type: "error", msg: err.message });
      handleCliError(err);
      process.exit(1);
    });
  }
);
